import os
from dataclasses import dataclass
from collections.abc import Sequence

import yaml
from duckdb import DuckDBPyConnection, CatalogException, ParserException, Error
from pyspark.sql import DataFrame, SparkSession

from databricks.labs.lakebridge.assessments.pipeline import PipelineClass

PROFILER_DB_NAME = "profiler_extract.db"


class SchemaDefinitionLoadError(Exception):
    """An exception that is raised when a schema definition cannot be loaded."""


class SchemaValidationError(Exception):
    """An exception that is raised when a schema cannot be validated against the source."""


@dataclass(frozen=True)
class ValidationOutcome:
    """A data class that holds the outcome of a table validation check."""

    table: str
    column: str | None
    strategy: str
    outcome: str
    severity: str
    summary: str | None


class ValidationStrategy:
    """Abstract class for validating a Profiler table"""

    def validate(self, connection: DuckDBPyConnection) -> ValidationOutcome:
        raise NotImplementedError


class NullValidationCheck(ValidationStrategy):
    """Concrete class for validating null values in a profiler table"""

    def __init__(self, table, column, severity="WARN"):
        self.name = self.__class__.__name__
        self.table = table
        self.column = column
        self.severity = severity

    def validate(self, connection: DuckDBPyConnection) -> ValidationOutcome:
        """
        Validates that a column does not contain null values.
        input:
          connection: a DuckDB connection object
        """
        result = connection.execute(f"SELECT COUNT(*) FROM {self.table} WHERE {self.column} IS NULL").fetchone()
        if result:
            row_count = result[0]
            outcome = "FAIL" if row_count > 0 else "PASS"
        else:
            outcome = "FAIL"
        return ValidationOutcome(self.table, self.column, self.name, outcome, self.severity, None)


class EmptyTableValidationCheck(ValidationStrategy):
    """Concrete class for validating empty tables from a profiler run."""

    def __init__(self, table, severity="WARN"):
        self.name = self.__class__.__name__
        self.table = table
        self.severity = severity

    def validate(self, connection) -> ValidationOutcome:
        """Validates that a table is not empty.
        input:
          connection: a DuckDB connection object
        returns:
          a ValidationOutcome object
        """
        try:
            result = connection.execute(f"SELECT COUNT(*) FROM {self.table}").fetchone()
            if result:
                row_count = result[0]
                outcome = "PASS" if row_count > 0 else "FAIL"
            else:
                outcome = "FAIL"
        except (CatalogException, ParserException):
            return ValidationOutcome(self.table, None, self.name, "FAIL", self.severity, "Table not found.")
        return ValidationOutcome(self.table, None, self.name, outcome, self.severity, None)


class ExtractSchemaValidationCheck(ValidationStrategy):
    """Concrete class for validating the schema of a profiler extract."""

    def __init__(
        self, schema: str, table: str, source_tech: str, extract_path: str, schema_path: str, severity: str = "WARN"
    ) -> None:
        self.name = self.__class__.__name__
        self.schema = schema
        self.table = table
        self.source_tech = source_tech
        self.extract_path = extract_path
        self.schema_path = schema_path
        self.severity = severity

    def _load_schema_definition(self) -> dict:
        """
        Loads a schema definition file from a local path.
        An `AssertionError` is raised if the schema does not match the expected source tech type.
        """
        try:
            with open(self.schema_path, "r", encoding="UTF-8") as f:
                schema_definition = yaml.safe_load(f)
        except FileNotFoundError as e:
            raise SchemaDefinitionLoadError(f"Schema definition file not found: '{self.schema_path}'") from e
        except (yaml.YAMLError, OSError) as e:
            raise SchemaDefinitionLoadError(f"Error while loading schema definition file: '{e}'") from e

        # Ensure that the correct schema definition was loaded
        assert (
            schema_definition["source_tech"].lower() == self.source_tech.lower()
        ), f"Incorrect schema definition type for source tech '{self.source_tech}'"

        return schema_definition

    def validate(self, connection) -> ValidationOutcome:
        """Validates that a table conforms to the expected schema.
        input:
          connection: a DuckDB connection object
        returns:
          a ValidationOutcome object
        """

        # First, load the table info from the schema definition YAML file
        schema_definition = self._load_schema_definition()

        # Load the table name and column info
        try:
            expected_columns = schema_definition["schemas"][self.schema]["tables"][self.table]["columns"]
        except KeyError as e:
            raise SchemaValidationError(
                f"Schema '{self.schema}' or table '{self.table}' could not be found in the schema definition."
            ) from e

        # Validate that:
        # a) the table exists in the profiler extract database
        # b) the columns for the table exist
        column_info_query = f"""
            SELECT column_name, data_type
              FROM information_schema.columns
             WHERE table_schema = '{self.schema}'
               AND table_name = '{self.table}'
        """
        try:
            result = connection.execute(column_info_query).fetchall()
        except (CatalogException, ParserException, Error) as e:
            raise SchemaValidationError(
                f"Could not query column information for table '{self.schema}.{self.table}' - {e}"
            ) from e

        # compare the extract schema with the schema definition
        if not result:
            return ValidationOutcome(
                f"{self.schema}.{self.table}",
                None,
                self.name,
                "FAIL",
                self.severity,
                "Table not found in the profiler extract",
            )

        extract_columns = dict(result)
        for col in expected_columns:
            expected_col_name = col["name"]
            expected_data_type = col["type"]

            # Column must be present in the extracted table
            if expected_col_name not in extract_columns:
                return ValidationOutcome(
                    f"{self.schema}.{self.table}",
                    expected_col_name,
                    self.name,
                    "FAIL",
                    self.severity,
                    "Column does not exist",
                )

            extracted_col_type = extract_columns[expected_col_name]
            if expected_data_type != extracted_col_type:
                return ValidationOutcome(
                    f"{self.schema}.{self.table}",
                    expected_col_name,
                    self.name,
                    "FAIL",
                    self.severity,
                    "Unexpected column data type",
                )

        return ValidationOutcome(
            f"{self.schema}.{self.table}", None, self.name, "PASS", self.severity, "All columns match expected schema"
        )


def get_profiler_extract_path(pipeline_config_path: str) -> str:
    """
    Returns the filesystem path of the profiler extract database.
    input:
       pipeline_config_path: the location of the pipeline definition .yml file
    returns:
       the filesystem path to the profiler extract database
    """
    pipeline_config = PipelineClass.load_config_from_yaml(pipeline_config_path)
    normalized_db_path = os.path.normpath(pipeline_config.extract_folder)
    database_path = f"{normalized_db_path}/{PROFILER_DB_NAME}"
    return database_path


def build_validation_report(
    validations: Sequence[ValidationStrategy], connection: DuckDBPyConnection
) -> list[ValidationOutcome]:
    """
    Builds a list of ValidationOutcomes from list of validation checks.
    input:
      validations: a list of ValidationStrategy objects
      connection: a DuckDB connection object
    returns: a list of ValidationOutcomes
    """
    validation_report = []
    for validation in validations:
        validation_report.append(validation.validate(connection))
    return validation_report


def build_validation_report_dataframe(
    validations: Sequence[ValidationStrategy], connection: DuckDBPyConnection
) -> DataFrame:
    """
    Builds a list of ValidationOutcomes from list of validation checks.
    input:
      validations: a list of ValidationStrategy objects
      connection: a DuckDB connection object
    returns: a list of ValidationOutcomes
    """
    validation_report = []
    for validation in validations:
        result = validation.validate(connection)
        validation_report.append(
            (result.table, result.column, result.strategy, result.outcome, result.severity, result.summary)
        )
    spark = SparkSession.builder.getOrCreate()
    schema = "table STRING, column STRING, strategy STRING, outcome STRING, severity STRING, summary STRING"
    return spark.createDataFrame(validation_report, schema)



import logging
import os
import sys
from pathlib import Path
import yaml
from yaml.parser import ParserError
from yaml.scanner import ScannerError

import duckdb
from pyspark.sql import SparkSession

from databricks.labs.lakebridge.assessments.profiler_validator import (
    EmptyTableValidationCheck,
    build_validation_report,
    ExtractSchemaValidationCheck,
    build_validation_report_dataframe,
)

logger = logging.getLogger(__name__)


def main(*argv) -> None:
    logger.debug(f"Arguments received: {argv}")
    assert len(sys.argv) == 4, f"Invalid number of arguments: {len(sys.argv)}"
    catalog_name = sys.argv[0]
    schema_name = sys.argv[1]
    extract_location = sys.argv[2]
    source_tech = sys.argv[3]
    logger.info(f"Validating {source_tech} profiler extract located at '{extract_location}'.")
    valid_extract = _validate_profiler_extract(catalog_name, schema_name, extract_location, source_tech)
    if valid_extract:
        _ingest_profiler_tables(catalog_name, schema_name, extract_location)
    else:
        raise ValueError("Corrupt or invalid profiler extract.")


def _get_extract_tables(schema_def_path: str) -> list:
    """
    Given a schema definition file for a source technology, returns a list of table info tuples:
    (schema_name, table_name, fully_qualified_name)
    """
    # First, load the schema definition file
    try:
        with open(schema_def_path, 'r', encoding="UTF-8") as f:
            data = yaml.safe_load(f)
    except (ParserError, ScannerError) as e:
        raise ValueError(f"Could not read extract schema definition '{schema_def_path}': {e}") from e
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Schema definition not found: {schema_def_path}") from e
    # Iterate through the defined schemas and build a list of
    # table info tuples: (schema_name, table_name, fully_qualified_name)
    extracted_tables = []
    for schema_name, schema_def in data.get("schemas", {}).items():
        tables = schema_def.get("tables", {})
        for table_name in tables.keys():
            fq_name = f"{schema_name}.{table_name}"
            extracted_tables.append((schema_name, table_name, fq_name))

    return extracted_tables


def _validate_profiler_extract(
    target_catalog_name: str, target_schema_name: str, extract_location: str, source_tech: str
) -> bool:
    logger.info("Validating the profiler extract file.")
    validation_checks: list[EmptyTableValidationCheck | ExtractSchemaValidationCheck] = []
    schema_def_path = f"{Path(__file__).parent}/../../resources/assessments/{source_tech}_schema_def.yml"
    tables = _get_extract_tables(schema_def_path)
    try:
        with duckdb.connect(database=extract_location) as duck_conn:
            for table_info in tables:
                # Ensure that the table contains data
                empty_check = EmptyTableValidationCheck(table_info[2])
                validation_checks.append(empty_check)

                # Ensure that the table conforms to the expected schema
                schema_check = ExtractSchemaValidationCheck(
                    table_info[0],
                    table_info[1],
                    source_tech=source_tech,
                    extract_path=extract_location,
                    schema_path=schema_def_path,
                )
                validation_checks.append(schema_check)
            report = build_validation_report(validation_checks, duck_conn)
    except duckdb.IOException as e:
        logger.exception(f"Could not access the profiler extract: '{extract_location}'.")
        raise e
    except Exception as e:
        logger.exception(f"Unable to validate the profiler extract: '{extract_location}'.")
        raise e

    # Save validation report to table
    report_df = build_validation_report_dataframe(validation_checks, duck_conn)
    validation_report_table = f"{target_catalog_name}.{target_schema_name}.validation_report"
    logger.info(f"Saving extract validation report to '{validation_report_table}' to Unity Catalog.")
    report_df.write.format("delta").mode("overwrite").saveAsTable(validation_report_table)

    if len(report) > 0:
        report_errors = list(filter(lambda x: x.outcome == "FAIL" and x.severity == "ERROR", report))
        num_errors = len(report_errors)
        logger.info(f"There are {num_errors} validation errors in the profiler extract.")

    else:
        raise ValueError("Profiler extract validation report is empty.")
    return num_errors == 0


def _ingest_profiler_tables(catalog_name: str, schema_name: str, extract_location: str) -> None:
    try:
        with duckdb.connect(database=extract_location) as duck_conn:
            tables_to_ingest = duck_conn.execute("SHOW ALL TABLES").fetchall()
    except duckdb.IOException as e:
        logger.error(f"Could not access the profiler extract: '{extract_location}': {e}")
        raise duckdb.IOException(f"Could not access the profiler extract: '{extract_location}'.") from e
    except Exception as e:
        logger.error(f"Unable to read tables from profiler extract: '{extract_location}': {e}")
        raise e

    if len(tables_to_ingest) == 0:
        raise ValueError("Profiler extract contains no tables.")

    successful_tables = []
    unsuccessful_tables = []
    for source_table in tables_to_ingest:
        try:
            fq_source_table_name = f"{source_table[0]}.{source_table[1]}.{source_table[2]}"
            fq_delta_table_name = f"{catalog_name}.{schema_name}.{source_table[2]}"
            logger.info(f"Ingesting profiler table: '{fq_source_table_name}'")
            _ingest_table(extract_location, fq_source_table_name, fq_delta_table_name)
            successful_tables.append(fq_source_table_name)
        except (ValueError, IndexError, TypeError) as e:
            logger.error(f"Failed to construct source and destination table names: {e}")
            unsuccessful_tables.append(source_table)
        except duckdb.Error as e:
            logger.error(f"Failed to ingest table from profiler database: {e}")
            unsuccessful_tables.append(source_table)
        except RuntimeError as e:
            logger.error(f"Unknown error while ingested table from profiler database: {e}")
            unsuccessful_tables.append(source_table)

    logger.info(f"Ingested {len(successful_tables)} tables from profiler extract.")
    logger.info(",".join(successful_tables))

    # Log failed tables if there were errors
    logger.warning(f"Failed to ingest {len(unsuccessful_tables)} tables from profiler extract.")
    logger.warning(",".join(str(t) for t in unsuccessful_tables))


def _ingest_table(extract_location: str, source_table_name: str, target_table_name: str) -> None:
    """
    Ingest a table from a DuckDB profiler extract into a managed Delta table in Unity Catalog.
    """
    try:
        with duckdb.connect(database=extract_location, read_only=True) as duck_conn:
            query = f"SELECT * FROM {source_table_name}"
            pdf = duck_conn.execute(query).df()
            # Save table as a managed Delta table in Unity Catalog
            logger.info(f"Saving profiler table '{target_table_name}' to Unity Catalog.")
            spark = SparkSession.builder.getOrCreate()
            df = spark.createDataFrame(pdf)
            df.write.format("delta").mode("overwrite").saveAsTable(target_table_name)
    except duckdb.CatalogException as e:
        logger.error(f"Could not find source table '{source_table_name}' in profiler extract: {e}")
        raise duckdb.CatalogException(f"Could not find source table '{source_table_name}' in profiler extract.") from e
    except duckdb.IOException as e:
        logger.error(f"Could not access the profiler extract: '{extract_location}': {e}")
        raise duckdb.IOException(f"Could not access the profiler extract: '{extract_location}'.") from e
    except Exception as e:
        logger.error(f"Unable to ingest table '{source_table_name}' from profiler extract: {e}")
        raise e


if __name__ == "__main__":
    # Ensure that the ingestion job is being run on a Databricks cluster
    if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
        raise SystemExit("The Lakebridge profiler ingestion job is only intended to run in a Databricks Runtime.")
    main(*sys.argv)
